# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SessionStateData(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'startup': 'SessionStartingUpData',
        'stopping': 'SessionStoppingData'
    }

    attribute_map = {
        'startup': 'startup',
        'stopping': 'stopping'
    }

    def __init__(self, startup=None, stopping=None, local_vars_configuration=None):  # noqa: E501
        """SessionStateData - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._startup = None
        self._stopping = None
        self.discriminator = None

        if startup is not None:
            self.startup = startup
        if stopping is not None:
            self.stopping = stopping

    @property
    def startup(self):
        """Gets the startup of this SessionStateData.  # noqa: E501


        :return: The startup of this SessionStateData.  # noqa: E501
        :rtype: SessionStartingUpData
        """
        return self._startup

    @startup.setter
    def startup(self, startup):
        """Sets the startup of this SessionStateData.


        :param startup: The startup of this SessionStateData.  # noqa: E501
        :type: SessionStartingUpData
        """

        self._startup = startup

    @property
    def stopping(self):
        """Gets the stopping of this SessionStateData.  # noqa: E501


        :return: The stopping of this SessionStateData.  # noqa: E501
        :rtype: SessionStoppingData
        """
        return self._stopping

    @stopping.setter
    def stopping(self, stopping):
        """Sets the stopping of this SessionStateData.


        :param stopping: The stopping of this SessionStateData.  # noqa: E501
        :type: SessionStoppingData
        """

        self._stopping = stopping

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionStateData):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionStateData):
            return True

        return self.to_dict() != other.to_dict()
